/*
  +----------------------------------------------------------------------+
  | Portable System Library                                              |
  | Author: Sukchan Lee (acetcom@postech.ac.kr)                          |
  +----------------------------------------------------------------------+

  $Id: memory.c 113 2004-04-06 06:02:40Z acetcom $
*/

#include "psys/common.h"
#include "psys/memory.h"
#include "psys/errno.h"

/* MIN_BLOCK_SIZE ּ ޸ Block ũ⸦ Ѵ. */
#define MIN_BLOCK_SIZE   8192
/* 
 * ޸ Block ũ⿡    Index .
 * size( 8kb - 12kb) ε 1 
 * size(12kb - 16kb) ε 2 
 * size(16kb - 20kb) ε 3 
 * ...
 * ...
 * ޸ Block ּ ũ 8kb̰, 4kb ũ⿡ 
 *   Index      ִ.
 *
 * ޸𸮸 Ҵ  MAX_BLOCK_INDEX ũ ǹ̰ .  ڰ
 * 84kb̻ ޸ Ҵ ûϸ  ޸ Block Index 20̻ 
 *   ȴ. ׷, ޸𸮸   Index 20̻ 
 * ͵鿡   0 Ͽ óѴ. 
 *
 * ᱹ, ޸ Ҵÿ ޸ Block Index 1̻   ǰ,
 * ޸ ÿ 0 19  ޸ Block Index ȴ.
 */
#define MAX_BLOCK_INDEX   20

/* [2^12 = 4096 = 4kb] ޸ Block Index   ̴. */
#define BOUNDARY_INDEX  12
#define BOUNDARY_SIZE   (1 << BOUNDARY_INDEX)

#define PSYS_BLOCK_T_SIZE       PSYS_ALIGN_DEFAULT(sizeof(psys_mem_block_t))
#define PSYS_CACHE_T_SIZE       PSYS_ALIGN_DEFAULT(sizeof(psys_mem_cache_t))
#define PSYS_POOL_T_SIZE        PSYS_ALIGN_DEFAULT(sizeof(psys_mem_pool_t))

typedef struct psys_mem_cache_t    psys_mem_cache_t;
typedef struct psys_mem_block_t    psys_mem_block_t;

/* 
 * Block Ҵ ޸ ҵ  ִ. ޸ Block 
 * Circular Linked List  ִµ, free_index ū  
 * SortingǾ ִ ̴. ٽ ؼ, ̴  ޸   
 * ū  Linked List Ǿ  ǹѴ.
 *
 * , ޸ Pool Ȱȭ(Active)¿ ִ ޸ Block 
 * Sorting ʴ´.
 */
struct psys_mem_block_t 
{
    /*  ޸ Block  */
    psys_mem_block_t    *next;  
    /*  ޸ Block next    */
    psys_mem_block_t    **ref;  
    /* ޸ Block  Index */
    p_uint32_t          index;  
    /* ϰ  ޸  Index */
    p_uint32_t          free_index;
    /*  ޸  ġ */
    p_int8_t            *first_avail;   
    /*  ޸  ġ */
    p_int8_t            *endp;
};

/*
 * Cache  ޸ Block  δ ̴.
 * , Pool ޸ Ҵ û ϸ Cache  ޸𸮷κ,
 *  ũ⸦  ޸ Block ִ  ˻Ѵ.
 *  Ҵ û ˸ ޸ Block , Cache Ǿ 
 * ִ ޸𸮸 Ҵϰ, ׷ , mallocԼ  OSκ 
 * ޸𸮸 ҴѴ.
 */
struct psys_mem_cache_t 
{
    /* 
     *  ޸ Block ߿  ū Index Ű ε,
     * ̴ Pool ޸ Ҵ ûÿ,   ޸ Block
     * Ҵ  ִ ˻ϴ  ̿ȴ. ̰ Index
     * ޸ Block ũ⸦ ϱ  ϴ.
     *
     *  , Index 5 ޸ Block Ͽ, Cache 
     *  ξٰ . ׷,  Cache ũⰡ
     * 24kb ޸ Block Ѵٴ  ϰ, max_index
     * 5 ̴.  ڰ 8kb ޸𸮿  Pool Ҵ 
     * û ϸ ̴ Index 1 ̹Ƿ max_index 񱳸 
     *  Cache ޸𸮷κ Ҵ   ִٴ  ǹѴ.
     * ׷, ڰ 32kb ޸ Ҵ û ϸ  Block Index
     * 7̹Ƿ max_index 񱳸  mallocԼ  OSκ 
     * ޸𸮸 Ҵ ־  ̴.
     */ 
    p_uint32_t          max_index;          
    /* 
     *  ޸ Block  δ ̴.
     * Index 20̻ ޸ Block   
     * free[0] ޸ Block  д.
     */
    psys_mem_block_t    *free[MAX_BLOCK_INDEX];   
};

/* 
 * ޸ Pool Ÿ üε, Hierarchy θ-ڽ 踦  ִ.
 * ̴ ޸ Pool  , ׿   ڽĵ ޸ Pool Բ 
 * ϴ   ȴ. libpsys  ޸ Pool 
 * θ-ڽİ 踸 ˸ Ǳ  Binary Tree  ʰ,
 * Left-Child, Right-Sibling Tree ̿Ͽ Ͽ.
 */ 
struct psys_mem_pool_t 
{
    psys_mem_pool_t         *parent;    /* θ Pool */
    psys_mem_pool_t         *child;     /* ڽ Pool */
    psys_mem_pool_t         *sibling;   /*  Pool */

    /* 
     * psys_pool_create ϸ psys_pool_t  ޸𸮵 ʿϴ.
     * self psys_pool_t ü  ޸ Block ϴ ̴.
     * 
     */
    psys_mem_block_t        *self;
    /*
     * psys_palloc ޸ Block ´. active 
     * Circular Linked List ޸ Pool ϴ 
     * ޸ Block ߿  Active ִ Block̴.
     *
     *  Active¿  ʴ ޸ Block active->next
     *  ޸ 翡  SortingǾ  ִ.
     */
    psys_mem_block_t        *active;
};

static p_uint8_t psys_mem_pool_initialized = 0;
static psys_mem_pool_t *psys_mem_pool = NULL;
static psys_mem_cache_t *psys_mem_cache = NULL;

psys_mem_block_t *psys_mem_cache_alloc(size_t size)
{
    psys_mem_block_t *block, **ref;
    p_uint32_t i, index, max_index;

    /* ޸ Block ּ ũ 8kb̴.*/
    size = PSYS_ALIGN(size + PSYS_BLOCK_T_SIZE, BOUNDARY_SIZE);
    if (size < MIN_BLOCK_SIZE)
    {
        size = MIN_BLOCK_SIZE;
    }

    /* 
     * Pool û ޸ Block  Index ϴ  Ѵ.
     * size( 8kb - 12kb) ε 1 
     * size(12kb - 16kb) ε 2 
     * size(16kb - 20kb) ε 3 
     * ...
     * ...
     */
    index = (size >> BOUNDARY_INDEX) - 1;

    /* 
     * ޸𸮸  , psys_mem_cache->max_index  Index 
     * ȴ.  , psys_psys_mem_cache_freeԼ  ޸ 
     * Block Ǹ, psys_mem_cache->max_index  index 
     *  д.  , psys_mem_cache->max_index  ޸ 
     * Block߿ ִ밪  index ϵ Ѵ.
     */
    if (index <= psys_mem_cache->max_index) 
    {
        /* 
         * ̹  ޸ Block  Ҵ ޸ Block ū ̴.
         * ٽ ؼ, ̹  ޸𸮿  ޸𸮸 Ҵ  ִٴ
         *  ǹѴ. Index ޸ Block ũ⸦ ǹϹǷ, ܼ 
         * Index Ͽ,  Ҵ ޸ Block  ޸ Block 
         * Ҵ   ˾Ƴ  ִ.
         */
        max_index = psys_mem_cache->max_index;
        ref = &psys_mem_cache->free[index];
        i = index;

        /*
         *  ޸ Block index max_index Ѵ.
         * ̴ ޸𸮸 Ҵϱ   ޸ Block߿   
         * Block ã  Ѵ. ̴ ޸ Ҵ ӵ  ϴ
         *  , ޸ ȿ  ϴ    ִ.
         */
        while (*ref == NULL && i < max_index) 
        {
           ref++;
           i++;
        }

        if ((block = *ref) != NULL) 
        {
            /*
             * ̹  ޸𸮸  ҴϷ  ,  
             * max_index ִ  ޸𸮶, ο max_index
             *  ξ Ѵ.
             */
            if ((*ref = block->next) == NULL && i >= max_index) 
            {
                do 
                {
                    ref--;
                    max_index--;
                }
                while (*ref == NULL && max_index > 0);

                psys_mem_cache->max_index = max_index;
            }

            block->next = NULL;
            block->first_avail = (char *)block + PSYS_BLOCK_T_SIZE;

            return block;
        }

    }
    else if (psys_mem_cache->free[0]) 
    {
        /*
         * ִ ޸ Block Index   ̹  ޸𸮰
         * ִ Ѵ.  Ѵٸ ⼭ ޸𸮸  ҴѴ.
         */
        ref = &psys_mem_cache->free[0];
        while ((block = *ref) != NULL && index > block->index)
        {
            ref = &block->next;
        }

        if (block) 
        {
            *ref = block->next;

            block->next = NULL;
            block->first_avail = (char *)block + PSYS_BLOCK_T_SIZE;

            return block;
        }
    }

    /*
     *  Ҵ ޸𸮰 ̹  ޸𸮿 ã  ٸ,
     * OSκ malloc ̿Ͽ ޸𸮸 ҴѴ.
     */
    if ((block = malloc(size)) == NULL)
    {
        return NULL;
    }

    block->next = NULL;
    block->index = index;
    block->first_avail = (char *)block + PSYS_BLOCK_T_SIZE;
    block->endp = (char *)block + size;

    return block;
}

void psys_mem_cache_free(psys_mem_block_t *block)
{
    psys_mem_block_t *next;
    p_uint32_t index, max_index;

    max_index = psys_mem_cache->max_index;

    do 
    {
        next = block->next;
        index = block->index;

        /*
         *  ޸𸮰 20  Ѵ.
         */
        if (index < MAX_BLOCK_INDEX) 
        {
            /*
             * 20  Index ؼ psys_mem_cache->free[index]
             *   ޸𸮸  д.
             */
            if ((block->next = psys_mem_cache->free[index]) == NULL && 
                    index > max_index) 
            {
                /*
                 *   ޸ Index  max_index ũ
                 * max_index  Ϸ ޸ Index UpdateѴ.
                 */
                max_index = index;
            }
            psys_mem_cache->free[index] = block;
        }
        else 
        {
            /*
             * Index 20̸̻ psys_mem_cache->free[0] 
             *  ޸𸮸  д.
             */
            block->next = psys_mem_cache->free[0];
            psys_mem_cache->free[0] = block;
        }
    } while ((block = next) != NULL);

    psys_mem_cache->max_index = max_index;
}

psys_status_t psys_mem_pool_create(
        psys_mem_pool_t **newpool, psys_mem_pool_t *parent)
{
    psys_mem_pool_t *pool;
    psys_mem_block_t *block;

    *newpool = NULL;

    if (!parent)
    {
        parent = psys_mem_pool;
    }

    if (psys_mem_cache == NULL)
    {
        return PSYS_NOT_INITIALIZED;
    }

    /* 
     * ޸ Pool  ޸ Ҵ Cacheκ ûѴ. 
     * , Cache  ޸ Block  OS mallocκ
     * ޸ Ҵ ûѴ.
     */
    if ((block = psys_mem_cache_alloc(
                    MIN_BLOCK_SIZE - PSYS_BLOCK_T_SIZE)) == NULL) 
    {
        return PSYS_NO_MEMORY;
    }

    block->next = block;
    block->ref = &block->next;

    pool = (psys_mem_pool_t *)block->first_avail;
    /*
     * Ҵ ޸ Block ޸ Pool ü ŭ
     * ̹ ߱   ޸  ġ ű.
     */
    block->first_avail = (char *)pool + PSYS_POOL_T_SIZE;

    pool->active = pool->self = block;
    pool->child = NULL;

    /* ޸ Pool Left-Child, Right-Sibling  Tree Ѵ. */
    if ((pool->parent = parent) != NULL) 
    {
        /* 
         *    ڽ , 
         *       ֵ Ѵ. ^^;
         */
        pool->sibling = parent->child;

        /* θ  ߿  ڽĸ Ų. */
        parent->child = pool;
    }
    else 
    {
        pool->sibling = NULL;
    }

    *newpool = pool;

    return PSYS_SUCCESS;
}

void psys_mem_pool_destroy(psys_mem_pool_t *pool)
{
    psys_mem_block_t *active;

    /* 
     *  ߿   Cache ȾƸԴ´. 
     * Pool   ޸𸮴 Cache ȴ.
     */
    while (pool->child)
    {
        psys_mem_pool_destroy(pool->child);
    }

    /* 
     *   ȸ ڽ θ Ű Ѵ.
     * ù° ڽ  ߿ Cache ȸ ȴ.
     */
    if (pool->parent) 
    {
        pool->parent->child = pool->sibling;
    }

    active = pool->self;
    *active->ref = NULL;

    /*
     * pool->self ޸ Pool  , Cacheκ Ҵ
     * ޸̴.  ޸𸮴 Cache  ̴.
     */
    psys_mem_cache_free(active);
}


psys_status_t psys_mem_pool_initialize(void)
{
    psys_status_t rv;

    if (psys_mem_pool_initialized)
    {
        return PSYS_SUCCESS;
    }

    if ((psys_mem_cache = malloc(PSYS_CACHE_T_SIZE)) == NULL)
    {
        return PSYS_NO_MEMORY;
    }

    memset(psys_mem_cache, 0, PSYS_CACHE_T_SIZE);

    if ((rv = psys_mem_pool_create(&psys_mem_pool, NULL)) != PSYS_SUCCESS) 
    {
        free(psys_mem_cache);
        psys_mem_cache = NULL;
        psys_mem_pool_initialized = 0;
        return rv;
    }

    psys_mem_pool_initialized++;

    return PSYS_SUCCESS;
}


void psys_mem_pool_terminate(void)
{
    p_uint32_t index;
    psys_mem_block_t *block, **ref;

    if (!psys_mem_pool_initialized)
    {
        return;
    }

    psys_mem_pool_destroy(psys_mem_pool);
    psys_mem_pool = NULL;

    for (index = 0; index < MAX_BLOCK_INDEX; index++) 
    {
        /* 
         * ̹  ޸𸮸 cache  ִٸ,
         * OS   ⼭ ̷.
         */
        ref = &psys_mem_cache->free[index];
        while ((block = *ref) != NULL) 
        {
            *ref = block->next;
            free(block);
        }
    }

    free(psys_mem_cache);
    psys_mem_cache = NULL;

    --psys_mem_pool_initialized;
}

void *psys_malloc(psys_mem_pool_t *pool, size_t size)
{
    psys_mem_block_t *active, *block;
    void *mem;
    p_uint32_t free_index;

    size = PSYS_ALIGN_DEFAULT(size);
    active = pool->active;

    /*
     * pool->active  ޸ Pool Active¿ ִ ޸ Block̴.
     * ڰ û ޸ ũⰡ  ޸ Block  ޸𸮿 
     * 밡ϸ,  Block ޸𸮸 Ѵ.
     */
    if (size < (size_t)(active->endp - active->first_avail)) 
    {
        mem = active->first_avail;
        active->first_avail += size;

        return mem;
    }

    /*
     * active active->next   ޸  active->next 
     * active Ƽ, ޸ Ҵ   ִ 찡   
     * ִ. ֳϸ, Sorting Inactive¿ ִ ޸ Block
     * Ǿ ֱ ̴. , active->next  ޸ 絵 
     *  ־ Ѵ.
     */
    block = active->next;
    if (size < (size_t)(block->endp - block->first_avail)) 
    {
        /*
         *  ° ޸ Block(Inactive¿ ִ ޸ Block ù°) 
         *  ޸  ϹǷ  Block ̿Ͽ ޸ Ҵ 
         * Ѵ. ̸   ° ޸𸮸  List ϴ ۾ 
         * Ѵ.
         */
        *block->ref = block->next;
        block->next->ref = block->ref;
    }
    else 
    {
        if ((block = psys_mem_cache_alloc(size)) == NULL) 
        {
            return NULL;
        }
    }

    block->free_index = 0;

    mem = block->first_avail;
    block->first_avail += size;

    /*
     * pool->active Circular Linked List̴. 
     *   ޸ Block List Head ߰Ѵ.
     */
    block->ref = active->ref;
    *block->ref = block;
    block->next = active;
    active->ref = &block->next;
    pool->active = block;

    /*
     *  ޸ Pool Ȱȭ Ǿ ִ ޸ Block  뷮 
     * Index Ѵ. Index ޸ ũ  
     * õ ġ̴.
     */
    free_index = (PSYS_ALIGN(active->endp - active->first_avail + 1,
                            BOUNDARY_SIZE) - BOUNDARY_SIZE) >> BOUNDARY_INDEX;
    active->free_index = free_index;

    /* 
     *   ޸ Block Active¿ ֱ  ״ ΰ,
     *  Active¿ ޸𸮴  Inactive· ٲ 
     *  ޸ 翡   ġ Sorting ־ Ѵ.
     */
    block = active->next;
    if (free_index >= block->free_index)
    {
        /*
         *  Inactive·  ޸ Block  ޸ 
         * ٸ ޸ Block麸 ũǷ Sorting Ѵ.
         */
        return mem;
    }

    /*
     *  ޸  ޸ Block SortingǾ ϹǷ,
     * Index Ͽ  Inactive·  ޸ Block ġ
     *  ߰  Ѵ.
     */
    do 
    {
        block = block->next;
    }
    while (free_index < block->free_index);

    /*  Inactive¿  ޸ Block List Ѵ. */
    *active->ref = active->next;
    active->next->ref = active->ref;

    /*
     *  Inactive¿  ޸ Block  ġ ߰ϸ,
     * Inactive¿ ִ ޸ Block Sorting ġ ȴ.
     */
    active->ref = block->ref;
    *active->ref = active;
    active->next = block;
    block->ref = &active->next;

    return mem;
}

void *psys_calloc(psys_mem_pool_t *pool, size_t size)
{
    void *mem;

    size = PSYS_ALIGN_DEFAULT(size);
    if ((mem = psys_malloc(pool, size)) != NULL) 
    {
        memset(mem, 0, size);
    }

    return mem;
}
