/*
  +----------------------------------------------------------------------+
  | Portable System Library                                              |
  | Author: Sukchan Lee (acetcom@postech.ac.kr)                          |
  +----------------------------------------------------------------------+

  $Id: string.c 107 2004-03-26 17:07:59Z acetcom $
*/

#include "psys/common.h"
#include "psys/errno.h"
#include "psys/string.h"

#include "CuTest.h"

#define PSYS_TEST_STRING "Thanks to My Wife, Sejin Kim"

extern psys_mem_pool_t *test_mem_pool;

static void strcasecmp_test(CuTest *tc)
{
    CuAssertTrue(tc, strcasecmp("FroboZZ", "frobozz") == 0);
    CuAssertTrue(tc, strcasecmp("frobozz", "frobozz") == 0);
    CuAssertTrue(tc, strcasecmp("frobozz", "FROBOZZ") == 0);
    CuAssertTrue(tc, strcasecmp("FROBOZZ", "froboz") != 0);
    CuAssertTrue(tc, strcasecmp("", "") == 0);
    CuAssertTrue(tc, strcasecmp("!#%&/()", "!#%&/()") == 0);
    CuAssertTrue(tc, strcasecmp("a", "b") < 0);
    CuAssertTrue(tc, strcasecmp("a", "B") < 0);
    CuAssertTrue(tc, strcasecmp("A", "b") < 0);
    CuAssertTrue(tc, strcasecmp("A", "B") < 0);
    CuAssertTrue(tc, strcasecmp("b", "a") > 0);
    CuAssertTrue(tc, strcasecmp("b", "A") > 0);
    CuAssertTrue(tc, strcasecmp("B", "a") > 0);
    CuAssertTrue(tc, strcasecmp("B", "A") > 0);
}

static void strncasecmp_test(CuTest *tc)
{
    CuAssertTrue(tc, strncasecmp("FroboZZ", "frobozK", 6) == 0);
    CuAssertTrue(tc, strncasecmp("frobozz", "frobozK", 7) != 0);
    CuAssertTrue(tc, strncasecmp("frobozz", "FROBOZK", 8) != 0);
    CuAssertTrue(tc, strncasecmp("", "", 100) == 0);
}

static void psys_strdup_test(CuTest *tc)
{
    char *string1;
    char *string2;
    char *string3;

    CuAssertTrue(tc, psys_strdup(test_mem_pool, NULL) == NULL);
    string1 = psys_strdup(test_mem_pool, PSYS_TEST_STRING);
    CuAssertTrue(tc, string1 != NULL);
    CuAssertStrEquals(tc, string1, PSYS_TEST_STRING);
    string2 = psys_strndup(test_mem_pool, PSYS_TEST_STRING, 5);
    CuAssertTrue(tc, strncmp(string2, PSYS_TEST_STRING, 5) == 0);
    string3 = psys_strndup(test_mem_pool, PSYS_TEST_STRING, 100);
    CuAssertTrue(tc, strncmp(string3, PSYS_TEST_STRING, 100) == 0);
}

static void snprintf_test(CuTest *tc)
{
    char buf[5];

    CuAssertTrue(tc, 3 == snprintf (buf, 0, "%s", "abc"));
    CuAssertTrue(tc, 3 == snprintf (NULL, 0, "%s", "abc"));
    CuAssertTrue(tc, 3 == snprintf (buf, 5, "%s", "abc"));
    CuAssertTrue(tc, 4 == snprintf (buf, 5, "%s", "abcd"));
    CuAssertTrue(tc, 9 == snprintf (buf, 5, "%s", "abcdefghi"));
}

static void strlcpy_test(CuTest *tc)
{
    char string[6];

    *(string + 5) = 'Z'; 

    *string = 'q';
    CuAssertTrue(tc, strlcpy(string, "" , 5) == 0);
    CuAssertTrue(tc, *string == '\0');
    *string = 'q';
    CuAssertTrue(tc, strlcpy(string, "abc" , 5) == 3);
    CuAssertTrue(tc, *(string + 3) == '\0');
    CuAssertStrEquals(tc, string, "abc");
    CuAssertTrue(tc, strlcpy(string, "abcd" , 5) == 4);
    CuAssertTrue(tc, *(string + 4) == '\0' );
    CuAssertTrue(tc, *(string + 5) == 'Z' );
    CuAssertStrEquals(tc, string, "abcd");
    CuAssertTrue(tc, strlcpy(string, "abcde" , 5) == 5);
    CuAssertTrue(tc, *(string + 4) == '\0' );
    CuAssertTrue(tc, *(string + 5) == 'Z' );
    CuAssertStrEquals(tc, string, "abcd");
    CuAssertTrue(tc, strlcpy(string, "abcdef" , 5) == 6);
    CuAssertTrue(tc, *(string + 4) == '\0' );
    CuAssertTrue(tc, *(string + 5) == 'Z' );
    CuAssertStrEquals(tc, string, "abcd");
    *string = 'Y';
    *(string + 1)= '\0';
    CuAssertTrue(tc, strlcpy(string, "Hello" , 0) == 5);
    CuAssertTrue(tc, *string == 'Y');
}

static void strlcat_test(CuTest *tc)
{
    char string[6];

    *(string + 5) = 'Z'; 

    *string = '\0';
    CuAssertTrue(tc, strlcat(string, "123" , 5) == 3);
    CuAssertTrue(tc, *(string + 3) == '\0' );
    CuAssertStrEquals(tc, string, "123");
    CuAssertTrue(tc, strlcat(string, "" , 5) == 3);
    CuAssertTrue(tc, *(string + 3) == '\0' );
    CuAssertStrEquals(tc, string, "123");
    CuAssertTrue(tc, strlcat(string, "4", 5) == 4);
    CuAssertStrEquals(tc, string, "1234");
    CuAssertTrue(tc, strlcat(string, "5", 5) == 5);
    CuAssertTrue(tc, *(string + 4) == '\0' );
    CuAssertStrEquals(tc, string, "1234");
    CuAssertTrue(tc, *(string + 5) == 'Z' );
    *string = 'Y';
    *(string + 1)= '\0';
    CuAssertTrue(tc, strlcat(string, "123" , 0) == 3);
    CuAssertTrue(tc, *string == 'Y');

    CuAssertTrue(tc, strlcpy(string, "hi", 5) == 2);
    CuAssertStrEquals(tc, string, "hi");
    CuAssertTrue(tc, strlcat(string, "t", 5) == 3);
    CuAssertStrEquals(tc, string, "hit");
}

static void snprintf_int64_test(CuTest *tc)
{
    char buf[100];

    p_int64_t i = PSYS_INT64_C(-42);
    p_uint64_t ui = PSYS_UINT64_C(42);
    p_uint64_t big = PSYS_UINT64_C(3141592653589793238);

    snprintf(buf, sizeof(buf), "%" PSYS_INT64_FORMAT, i);
    CuAssertStrEquals(tc, buf, "-42");
    CuAssertTrue(tc, psys_atoi64(buf) == i);

    snprintf(buf, sizeof(buf), "%" PSYS_UINT64_FORMAT, ui);
    CuAssertStrEquals(tc, buf, "42");
    CuAssertTrue(tc, psys_atoi64(buf) == ui);

    snprintf(buf, sizeof(buf), "%" PSYS_UINT64_FORMAT, big);
    CuAssertStrEquals(tc, buf, "3141592653589793238");
    CuAssertTrue(tc, psys_atoi64(buf) == big);
}


static void psys_strtoi64_test(CuTest *tc)
{
    char *string = psys_strdup(test_mem_pool, "0x1d636b02300a7aa7");
    CuAssertTrue(tc, psys_atoi64(string) == PSYS_UINT64_C(0x1d636b02300a7aa7));
}

static void psys_itoa_test(CuTest *tc)
{
    char *string1 = psys_itoa(test_mem_pool, -42);
    char *string2 = psys_ltoa(test_mem_pool, 42);
    CuAssertStrEquals(tc, string1, "-42");
    CuAssertStrEquals(tc, string2, "42");
}

CuSuite *psys_string_test(void)
{
    CuSuite *suite = CuSuiteNew("String");

    SUITE_ADD_TEST(suite, strcasecmp_test);
    SUITE_ADD_TEST(suite, strncasecmp_test);
    SUITE_ADD_TEST(suite, psys_strdup_test);
    SUITE_ADD_TEST(suite, snprintf_test);
    SUITE_ADD_TEST(suite, strlcpy_test);
    SUITE_ADD_TEST(suite, strlcat_test);
    SUITE_ADD_TEST(suite, snprintf_int64_test);
    SUITE_ADD_TEST(suite, psys_strtoi64_test);
    SUITE_ADD_TEST(suite, psys_itoa_test);

    return suite;
}
